from collections import namedtuple

from dagster import check

from .context import PipelineContextDefinition

from .dependency import (
    DependencyDefinition,
    Solid,
    SolidInputHandle,
    SolidInstance,
    SolidOutputHandle,
)

from .expectation import ExpectationDefinition, ExpectationResult

from .infos import ContextCreationExecutionInfo, ExpectationExecutionInfo, TransformExecutionInfo

from .input import InputDefinition

from .output import OutputDefinition

from .resource import ResourceDefinition

from .repository import RepositoryDefinition

from .pipeline import PipelineDefinition, solids_in_topological_order

from .pipeline_creation import create_execution_structure

from .solid import SolidDefinition

from .utils import DEFAULT_OUTPUT


class Result(namedtuple('_Result', 'value output_name')):
    '''A solid transform function return a stream of Result objects.
    An implementator of a SolidDefinition must provide a transform that
    yields objects of this type.

    Attributes:
        value (Any): Value returned by the transform.
        output_name (str): Name of the output returns. defaults to "result"
'''

    def __new__(cls, value, output_name=DEFAULT_OUTPUT):
        return super(Result, cls).__new__(cls, value, check.str_param(output_name, 'output_name'))



from dagster import PipelineDefinition, execute_pipeline, lambda_solid


@lambda_solid
def hello_world():
    return 'hello'


def define_hello_world_pipeline():
    return PipelineDefinition(
        name='hello_world_pipeline', solids=[hello_world]
    )


if __name__ == '__main__':
    result = execute_pipeline(define_hello_world_pipeline())
    assert result.success



import pytest

from dagster import execute_pipeline, PipelineConfigEvaluationError
from dagster.tutorials.intro_tutorial.configuration_schemas import (
    define_demo_configuration_schema_pipeline,
    define_demo_configuration_schema_repo,
    define_typed_demo_configuration_schema_pipeline,
)
from dagster.utils import script_relative_path
from dagster.utils.yaml_utils import load_yaml_from_path


def intro_tutorial_path(path):
    return script_relative_path('../../../dagster/tutorials/intro_tutorial/{}'.format(path))


def test_demo_configuration_schema_pipeline_correct_yaml():
    result = execute_pipeline(
        define_demo_configuration_schema_pipeline(),
        load_yaml_from_path(intro_tutorial_path('configuration_schemas.yml')),
    )
    assert result.success
    assert len(result.result_list) == 2
    count_letters_result = result.result_for_solid('count_letters').transformed_value()
    expected_value = {'q': 2, 'u': 4, 'x': 2}
    assert set(count_letters_result.keys()) == set(expected_value.keys())
    for key, value in expected_value.items():
        assert count_letters_result[key] == value
    assert result.result_for_solid('multiply_the_word').transformed_value() == 'quuxquux'


def test_demo_configuration_schema_pipeline_runtime_error():
    with pytest.raises(TypeError):
        execute_pipeline(
            define_demo_configuration_schema_pipeline(),
            load_yaml_from_path(intro_tutorial_path('configuration_schemas_runtime_error.yml')),
        )


def test_demo_configuration_schema_pipeline_wrong_field():
    with pytest.raises(
        PipelineConfigEvaluationError,
        match=('Undefined field "multiply_the_word_with_typed_config" at path ' 'root:solids'),
    ):
        execute_pipeline(
            define_demo_configuration_schema_pipeline(),
            load_yaml_from_path(intro_tutorial_path('configuration_schemas_wrong_field.yml')),
        )


def test_typed_demo_configuration_schema_pipeline_correct_yaml():
    result = execute_pipeline(
        define_typed_demo_configuration_schema_pipeline(),
        load_yaml_from_path(intro_tutorial_path('configuration_schemas_typed.yml')),
    )
    assert result.success
    assert len(result.result_list) == 2
    count_letters_result = result.result_for_solid('count_letters').transformed_value()
    expected_value = {'q': 2, 'u': 4, 'x': 2}
    assert set(count_letters_result.keys()) == set(expected_value.keys())
    for key, value in expected_value.items():
        assert count_letters_result[key] == value
    assert result.result_for_solid('typed_multiply_the_word').transformed_value() == 'quuxquux'


def test_typed_demo_configuration_schema_type_mismatch_error():
    with pytest.raises(
        PipelineConfigEvaluationError,
        match=(
            'Type failure at path "root:solids:typed_multiply_the_word:config:factor" on type '
            '"Int". Got "\'not_a_number\'"'
        ),
    ):
        execute_pipeline(
            define_typed_demo_configuration_schema_pipeline(),
            load_yaml_from_path(
                script_relative_path(
                    (
                        '../../../dagster/tutorials/intro_tutorial/'
                        'configuration_schemas_type_mismatch_error.yml'
                    )
                )
            ),
        )


def test_configuration_schema_repository():
    repo = define_demo_configuration_schema_repo()
    assert repo
    assert repo.get_all_pipelines()



# encoding: utf-8

from dagster import execute_pipeline
from dagster.tutorials.intro_tutorial.config import (
    define_configurable_hello_pipeline,
    test_intro_tutorial_part_four,
)
from dagster.tutorials.utils import check_cli_execute_file_pipeline, check_script
from dagster.utils import script_relative_path


def test_tutorial_part_four():
    pipeline = define_configurable_hello_pipeline()

    result = execute_pipeline(pipeline, {'solids': {'configurable_hello': {'config': 'cn'}}})

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_for_solid('configurable_hello').transformed_value() == '你好, 世界!'

    result = execute_pipeline(pipeline, {'solids': {'configurable_hello': {'config': 'haw'}}})

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_for_solid('configurable_hello').transformed_value() == 'Aloha honua!'

    result = execute_pipeline(pipeline, {'solids': {'configurable_hello': {'config': 'es'}}})

    assert result.success
    assert len(result.result_list) == 1
    assert result.result_for_solid('configurable_hello').transformed_value() == 'Hello, world!'


def test_tutorial_script_part_four():
    check_script(script_relative_path('../../../dagster/tutorials/intro_tutorial/config.py'))


def test_tutorial_cli_part_four():
    check_cli_execute_file_pipeline(
        script_relative_path('../../../dagster/tutorials/intro_tutorial/config.py'),
        'define_configurable_hello_pipeline',
        script_relative_path('../../../dagster/tutorials/intro_tutorial/config_env.yml'),
    )


def test_test_intro_tutorial_part_four():
    test_intro_tutorial_part_four()



from __future__ import absolute_import
from collections import namedtuple
import copy
import json
import logging
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
import traceback

import coloredlogs

from dagster import check

VALID_LEVELS = set([CRITICAL, DEBUG, ERROR, INFO, WARNING])

LOOKUP = {'CRITICAL': CRITICAL, 'DEBUG': DEBUG, 'ERROR': ERROR, 'INFO': INFO, 'WARNING': WARNING}


def level_from_string(string):
    check.str_param(string, 'string')
    return LOOKUP[string]


class CompositeLogger(object):
    def __init__(self, loggers=None):
        self.loggers = check.opt_list_param(loggers, 'loggers', of_type=logging.Logger)

    def __getattr__(self, name):
        def _invoke_logger_method(*args, **kwargs):
            for logger in self.loggers:
                logger_method = check.is_callable(getattr(logger, name))
                logger_method(*args, **kwargs)

        return _invoke_logger_method


class JsonFileHandler(logging.Handler):
    def __init__(self, json_path):
        super(JsonFileHandler, self).__init__()
        self.json_path = check.str_param(json_path, 'json_path')

    def emit(self, record):
        try:
            log_dict = copy.copy(record.__dict__)

            # This horrific monstrosity is to maintain backwards compatability
            # with the old behavior of the JsonFileHandler, which the clarify
            # project has a dependency on. It relied on the dagster-defined
            # properties smashing all the properties of the LogRecord object
            # and uploads all of those properties to a redshift table for
            # in order to do analytics on the log

            if 'dagster_meta' in log_dict:
                dagster_meta_dict = log_dict['dagster_meta']
                del log_dict['dagster_meta']
            else:
                dagster_meta_dict = {}

            log_dict.update(dagster_meta_dict)

            with open(self.json_path, 'a') as ff:
                text_line = json.dumps(log_dict)
                ff.write(text_line + '\n')
        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('Error during logging!')
            logging.exception(str(e))


class StructuredLoggerMessage(
    namedtuple('_StructuredLoggerMessage', 'name message level meta record')
):
    def __new__(cls, name, message, level, meta, record):
        return super(StructuredLoggerMessage, cls).__new__(
            cls,
            check.str_param(name, 'name'),
            check.str_param(message, 'message'),
            check_valid_level_param(level),
            check.dict_param(meta, 'meta'),
            check.inst_param(record, 'record', logging.LogRecord),
        )


class JsonEventLoggerHandler(logging.Handler):
    def __init__(self, json_path, construct_event_record):
        super(JsonEventLoggerHandler, self).__init__()
        self.json_path = check.str_param(json_path, 'json_path')
        self.construct_event_record = construct_event_record

    def emit(self, record):
        try:
            event_record = self.construct_event_record(record)
            with open(self.json_path, 'a') as ff:
                text_line = json.dumps(event_record.to_dict())
                ff.write(text_line + '\n')

        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('Error during logging!')
            logging.exception(str(e))


class StructuredLoggerHandler(logging.Handler):
    def __init__(self, callback):
        super(StructuredLoggerHandler, self).__init__()
        self.callback = check.is_callable(callback, 'callback')

    def emit(self, record):
        try:
            self.callback(
                StructuredLoggerMessage(
                    name=record.name,
                    message=record.msg,
                    level=record.levelno,
                    meta=record.dagster_meta,
                    record=record,
                )
            )
        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('Error during logging!')
            logging.exception(str(e))


def check_valid_level_param(level):
    check.param_invariant(
        level in VALID_LEVELS,
        'level',
        'Must be valid python logging level. Got {level}'.format(level=level),
    )
    return level


def construct_single_handler_logger(name, level, handler):
    check.str_param(name, 'name')
    check_valid_level_param(level)
    check.inst_param(handler, 'handler', logging.Handler)

    klass = logging.getLoggerClass()
    logger = klass(name, level=level)
    logger.addHandler(handler)
    return logger


def define_structured_logger(name, callback, level):
    check.str_param(name, 'name')
    check.callable_param(callback, 'callback')
    check_valid_level_param(level)

    return construct_single_handler_logger(name, level, StructuredLoggerHandler(callback))


def define_json_file_logger(name, json_path, level):
    check.str_param(name, 'name')
    check.str_param(json_path, 'json_path')
    check_valid_level_param(level)

    stream_handler = JsonFileHandler(json_path)
    stream_handler.setFormatter(define_default_formatter())
    return construct_single_handler_logger(name, level, stream_handler)


def define_colored_console_logger(name, level=INFO):
    check.str_param(name, 'name')
    check.param_invariant(
        level in VALID_LEVELS,
        'level',
        'Must be valid python logging level. Got {level}'.format(level=level),
    )

    klass = logging.getLoggerClass()
    logger = klass(name, level=level)
    coloredlogs.install(logger=logger, level=level, fmt=default_format_string())
    return logger


def default_format_string():
    return '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


def define_default_formatter():
    return logging.Formatter(default_format_string())


def debug_format_string():
    return '''%(name)s.%(levelname)s: %(message)s
    time: %(asctime)s relative: %(relativeCreated)dms
    path: %(pathname)s line: %(lineno)d'''


def define_debug_formatter():
    return logging.Formatter(debug_format_string())


def get_formatted_stack_trace(exception):
    check.inst_param(exception, 'exception', Exception)
    if hasattr(exception, '__traceback__'):
        tb = exception.__traceback__
    else:
        import sys

        _exc_type, _exc_value, tb = sys.exc_info()
    return ''.join(traceback.format_tb(tb))



import contextlib
import errno
import inspect
import os
import re
import yaml

from dagster import check

from .yaml_utils import load_yaml_from_path, load_yaml_from_globs, load_yaml_from_glob_list


def script_relative_path(file_path):
    '''
    Useful for testing with local files. Use a path relative to where the
    test resides and this function will return the absolute path
    of that file. Otherwise it will be relative to script that
    ran the test
    '''
    # from http://bit.ly/2snyC6s

    check.str_param(file_path, 'file_path')
    scriptdir = inspect.stack()[1][1]
    return os.path.abspath(os.path.join(os.path.dirname(os.path.abspath(scriptdir)), file_path))


# Adapted from https://github.com/okunishinishi/python-stringcase/blob/master/stringcase.py
def camelcase(string):
    check.str_param(string, 'string')

    string = re.sub(r'^[\-_\.]', '', str(string))
    if not string:
        return string
    return str(string[0]).upper() + re.sub(
        r'[\-_\.\s]([a-z])', lambda matched: str(matched.group(1)).upper(), string[1:]
    )


def single_item(ddict):
    check.dict_param(ddict, 'ddict')
    check.param_invariant(len(ddict) == 1, 'ddict')
    return list(ddict.items())[0]


@contextlib.contextmanager
def pushd(path):
    old_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(old_cwd)


def safe_isfile(path):
    '''"Backport of Python 3.8 os.path.isfile behavior.

    This is intended to backport https://docs.python.org/dev/whatsnew/3.8.html#os-path. I'm not
    sure that there are other ways to provoke this behavior on Unix other than the null byte,
    but there are certainly other ways to do it on Windows. Afaict, we won't mask other
    ValueErrors, and the behavior in the status quo ante is rough because we risk throwing an
    unexpected, uncaught ValueError from very deep in our logic.
    '''
    try:
        return os.path.isfile(path)
    except ValueError:
        return False


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def merge_dicts(left, right):
    check.dict_param(left, 'left')
    check.dict_param(right, 'right')

    result = left.copy()
    result.update(right)
    return result



from dagster import (
    DependencyDefinition,
    InputDefinition,
    Int,
    OutputDefinition,
    PipelineDefinition,
    lambda_solid,
)
from dagster.core.execution import (
    execute_externalized_plan,
    create_execution_plan,
    ExecutionMetadata,
)
from dagster.core.execution_plan.objects import StepKind
from dagster.core.types.runtime import resolve_to_runtime_type

from dagster.core.types.marshal import serialize_to_file, deserialize_from_file

from dagster.utils.test import get_temp_file_names


def test_basic_pipeline_external_plan_execution():
    @lambda_solid
    def return_one():
        return 1

    @lambda_solid(inputs=[InputDefinition('num', Int)], output=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    pipeline = PipelineDefinition(
        name='basic_external_plan_execution',
        solids=[return_one, add_one],
        dependencies={'add_one': {'num': DependencyDefinition('return_one')}},
    )

    with get_temp_file_names(2) as temp_files:

        temp_path, write_path = temp_files  # pylint: disable=W0632

        int_type = resolve_to_runtime_type(Int)

        serialize_to_file(int_type.serialization_strategy, 5, temp_path)

        execution_plan = create_execution_plan(pipeline)

        results = execute_externalized_plan(
            pipeline,
            execution_plan,
            ['add_one.transform'],
            inputs_to_marshal={'add_one.transform': {'num': temp_path}},
            outputs_to_marshal={'add_one.transform': [{'output': 'result', 'path': write_path}]},
            execution_metadata=ExecutionMetadata(),
        )

        assert deserialize_from_file(int_type.serialization_strategy, write_path) == 6

    assert len(results) == 2

    thunk_step_result = results[0]

    assert thunk_step_result.kind == StepKind.VALUE_THUNK

    transform_step_result = results[1]
    assert transform_step_result.kind == StepKind.TRANSFORM
    assert transform_step_result.success
    assert transform_step_result.success_data.output_name == 'result'
    assert transform_step_result.success_data.value == 6



import click

from ..version import __version__
from .pipeline import create_pipeline_cli


def create_dagster_cli():
    @click.group(commands={'pipeline': create_pipeline_cli()})
    @click.version_option(version=__version__)
    def group():
        pass

    return group


def main():
    cli = create_dagster_cli()
    cli(obj={})  # pylint:disable=E1123



import pytest

from dagster import (
    ExecutionContext,
    ExecutionMetadata,
    PipelineContextDefinition,
    PipelineDefinition,
    check,
    lambda_solid,
)

from dagster.core.system_config.objects import EnvironmentConfig

from dagster.core.execution import create_execution_plan

from dagster.core.execution_plan.create import (
    ExecutionPlanInfo,
    create_execution_plan_core,
    create_execution_plan_from_steps,
)

from dagster.core.execution_plan.objects import ExecutionStep, StepKind, StepBuilderState

from dagster.core.execution_plan.simple_engine import execute_step

from dagster.utils.test import create_test_runtime_execution_context


def silencing_default_context():
    return {'default': PipelineContextDefinition(context_fn=lambda *_args: ExecutionContext())}


@lambda_solid
def noop():
    return 'foo'


def silencing_pipeline(solids):
    return PipelineDefinition(solids=solids, context_definitions=silencing_default_context())


def test_compute_noop_node_core():
    pipeline = silencing_pipeline(solids=[noop])

    environment = EnvironmentConfig()

    plan = create_execution_plan_core(
        ExecutionPlanInfo(create_test_runtime_execution_context(), pipeline, environment),
        ExecutionMetadata(),
    )

    assert len(plan.steps) == 1

    outputs = list(execute_step(plan.steps[0], create_test_runtime_execution_context(), {}))

    assert outputs[0].success_data.value == 'foo'


def test_compute_noop_node():
    pipeline = silencing_pipeline(solids=[noop])

    plan = create_execution_plan(pipeline)

    assert len(plan.steps) == 1
    outputs = list(execute_step(plan.steps[0], create_test_runtime_execution_context(), {}))

    assert outputs[0].success_data.value == 'foo'


def test_duplicate_steps():
    @lambda_solid
    def foo():
        pass

    with pytest.raises(check.CheckError):
        create_execution_plan_from_steps(
            [
                ExecutionStep(
                    'same_name',
                    [],
                    [],
                    lambda *args, **kwargs: None,
                    StepKind.TRANSFORM,
                    foo,
                    StepBuilderState(
                        pipeline_name='dummy',
                        initial_tags=dict(
                            pipeline='dummy', solid='dummy', solid_definition='dummy'
                        ),
                    ).get_tags(),
                ),
                ExecutionStep(
                    'same_name',
                    [],
                    [],
                    lambda *args, **kwargs: None,
                    StepKind.TRANSFORM,
                    foo,
                    StepBuilderState(
                        pipeline_name='dummy',
                        initial_tags=dict(
                            pipeline='dummy', solid='dummy', solid_definition='dummy'
                        ),
                    ).get_tags(),
                ),
            ]
        )



from collections import namedtuple
from contextlib import contextmanager
import copy
from enum import Enum
import toposort

from dagster import check
from dagster.utils import merge_dicts
from dagster.core.system_config.objects import EnvironmentConfig
from dagster.core.definitions import PipelineDefinition, Solid, SolidOutputHandle
from dagster.core.errors import DagsterError
from dagster.core.execution_context import RuntimeExecutionContext
from dagster.core.types.runtime import RuntimeType


class StepOutputHandle(namedtuple('_StepOutputHandle', 'step output_name')):
    def __new__(cls, step, output_name):
        return super(StepOutputHandle, cls).__new__(
            cls,
            step=check.inst_param(step, 'step', ExecutionStep),
            output_name=check.str_param(output_name, 'output_name'),
        )

    # Make this hashable so it be a key in a dictionary

    def __str__(self):
        return 'StepOutputHandle' '(step="{step.key}", output_name="{output_name}")'.format(
            step=self.step, output_name=self.output_name
        )

    def __repr__(self):
        return 'StepOutputHandle' '(step="{step.key}", output_name="{output_name}")'.format(
            step=self.step, output_name=self.output_name
        )

    def __hash__(self):
        return hash(self.step.key + self.output_name)

    def __eq__(self, other):
        return self.step.key == other.step.key and self.output_name == other.output_name


class StepSuccessData(namedtuple('_StepSuccessData', 'output_name value')):
    def __new__(cls, output_name, value):
        return super(StepSuccessData, cls).__new__(
            cls, output_name=check.str_param(output_name, 'output_name'), value=value
        )


class StepFailureData(namedtuple('_StepFailureData', 'dagster_error')):
    def __new__(cls, dagster_error):
        return super(StepFailureData, cls).__new__(
            cls, dagster_error=check.inst_param(dagster_error, 'dagster_error', DagsterError)
        )


class StepResult(namedtuple('_StepResult', 'success step kind success_data failure_data')):
    @staticmethod
    def success_result(step, kind, success_data):
        return StepResult(
            success=True,
            step=check.inst_param(step, 'step', ExecutionStep),
            kind=check.inst_param(kind, 'kind', StepKind),
            success_data=check.inst_param(success_data, 'success_data', StepSuccessData),
            failure_data=None,
        )

    @staticmethod
    def failure_result(step, kind, failure_data):
        return StepResult(
            success=False,
            step=check.inst_param(step, 'step', ExecutionStep),
            kind=check.inst_param(kind, 'kind', StepKind),
            success_data=None,
            failure_data=check.inst_param(failure_data, 'failure_data', StepFailureData),
        )


class StepKind(Enum):
    TRANSFORM = 'TRANSFORM'
    INPUT_EXPECTATION = 'INPUT_EXPECTATION'
    OUTPUT_EXPECTATION = 'OUTPUT_EXPECTATION'
    JOIN = 'JOIN'
    SERIALIZE = 'SERIALIZE'
    INPUT_THUNK = 'INPUT_THUNK'
    MATERIALIZATION_THUNK = 'MATERIALIZATION_THUNK'
    VALUE_THUNK = 'VALUE_THUNK'


class StepInput(namedtuple('_StepInput', 'name runtime_type prev_output_handle')):
    def __new__(cls, name, runtime_type, prev_output_handle):
        return super(StepInput, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            runtime_type=check.inst_param(runtime_type, 'runtime_type', RuntimeType),
            prev_output_handle=check.inst_param(
                prev_output_handle, 'prev_output_handle', StepOutputHandle
            ),
        )


class StepOutput(namedtuple('_StepOutput', 'name runtime_type')):
    def __new__(cls, name, runtime_type):
        return super(StepOutput, cls).__new__(
            cls,
            name=check.str_param(name, 'name'),
            runtime_type=check.inst_param(runtime_type, 'runtime_type', RuntimeType),
        )


class ExecutionStep(
    namedtuple(
        '_ExecutionStep',
        'key step_inputs step_input_dict step_outputs step_output_dict compute_fn kind solid tags',
    )
):
    def __new__(cls, key, step_inputs, step_outputs, compute_fn, kind, solid, tags):
        check.param_invariant('pipeline' in tags, 'tags', 'Step must have pipeline tag')
        check.param_invariant('solid' in tags, 'tags', 'Step must have solid tag')
        check.param_invariant('solid_definition' in tags, 'tags', 'Step must have solid tag')
        check.dict_param(tags, 'tags', key_type=str, value_type=str)

        return super(ExecutionStep, cls).__new__(
            cls,
            key=check.str_param(key, 'key'),
            step_inputs=check.list_param(step_inputs, 'step_inputs', of_type=StepInput),
            step_input_dict={si.name: si for si in step_inputs},
            step_outputs=check.list_param(step_outputs, 'step_outputs', of_type=StepOutput),
            step_output_dict={so.name: so for so in step_outputs},
            compute_fn=check.callable_param(compute_fn, 'compute_fn'),
            kind=check.inst_param(kind, 'kind', StepKind),
            solid=check.inst_param(solid, 'solid', Solid),
            tags=merge_dicts({'step_key': key}, tags),
        )

    @property
    def pipeline_name(self):
        return self.tags['pipeline']

    @property
    def solid_name(self):
        return self.tags['solid']

    @property
    def solid_definition_name(self):
        return self.tags['solid_definition']

    def __getnewargs__(self):
        # print('getnewargs was called')
        return (
            self.key,
            self.step_inputs,
            self.step_outputs,
            self.compute_fn,
            self.kind,
            self.solid,
        )

    def with_new_inputs(self, step_inputs):
        return ExecutionStep(
            key=self.key,
            step_inputs=step_inputs,
            step_outputs=self.step_outputs,
            compute_fn=self.compute_fn,
            kind=self.kind,
            solid=self.solid,
            tags=self.tags,
        )

    def has_step_output(self, name):
        check.str_param(name, 'name')
        return name in self.step_output_dict

    def step_output_named(self, name):
        check.str_param(name, 'name')
        return self.step_output_dict[name]

    def has_step_input(self, name):
        check.str_param(name, 'name')
        return name in self.step_input_dict

    def step_input_named(self, name):
        check.str_param(name, 'name')
        return self.step_input_dict[name]


class ExecutionValueSubplan(
    namedtuple('ExecutionValueSubplan', 'steps terminal_step_output_handle')
):
    '''
    A frequent pattern in the execution engine is to take a single value
    (e.g. an input or an output of a transform) and then flow that value
    value through a sequence of system-injected steps (e.g. expectations
    or materializations). This object captures that pattern. It contains
    all of the steps that comprise that Subplan and also a single output
    handle that points to output that further steps down the plan can
    depend on.
    '''

    def __new__(cls, steps, terminal_step_output_handle):
        return super(ExecutionValueSubplan, cls).__new__(
            cls,
            check.list_param(steps, 'steps', of_type=ExecutionStep),
            check.inst_param(
                terminal_step_output_handle, 'terminal_step_output_handle', StepOutputHandle
            ),
        )

    @staticmethod
    def empty(terminal_step_output_handle):
        return ExecutionValueSubplan([], terminal_step_output_handle)


class ExecutionPlan(object):
    def __init__(self, step_dict, deps):
        self.step_dict = check.dict_param(
            step_dict, 'step_dict', key_type=str, value_type=ExecutionStep
        )
        self.deps = check.dict_param(deps, 'deps', key_type=str, value_type=set)
        self.steps = list(step_dict.values())

    def has_step(self, key):
        check.str_param(key, 'key')
        return key in self.step_dict

    def get_step_by_key(self, key):
        check.str_param(key, 'key')
        return self.step_dict[key]

    def topological_steps(self):
        return list(self._topological_steps())

    def _topological_steps(self):
        ordered_step_keys = toposort.toposort_flatten(self.deps)
        for step_key in ordered_step_keys:
            yield self.step_dict[step_key]


class ExecutionPlanInfo(namedtuple('_ExecutionPlanInfo', 'context pipeline environment')):
    def __new__(cls, context, pipeline, environment):
        return super(ExecutionPlanInfo, cls).__new__(
            cls,
            check.inst_param(context, 'context', RuntimeExecutionContext),
            check.inst_param(pipeline, 'pipeline', PipelineDefinition),
            check.inst_param(environment, 'environment', EnvironmentConfig),
        )


class ExecutionPlanSubsetInfo(namedtuple('_ExecutionPlanSubsetInfo', 'subset inputs')):
    '''
    inputs is a two dimensional dictionary that maps step_key => input_name => input_value
    '''

    def __new__(cls, included_steps, inputs=None):
        return super(ExecutionPlanSubsetInfo, cls).__new__(
            cls,
            set(check.list_param(included_steps, 'included_steps', of_type=str)),
            check.opt_dict_param(inputs, 'inputs'),
        )


class StepOutputMap(dict):
    def __getitem__(self, key):
        check.inst_param(key, 'key', SolidOutputHandle)
        return dict.__getitem__(self, key)

    def __setitem__(self, key, val):
        check.inst_param(key, 'key', SolidOutputHandle)
        check.inst_param(val, 'val', StepOutputHandle)
        return dict.__setitem__(self, key, val)


# This is the state that is built up during the execution plan build process.
# steps is just a list of the steps that have been created
# step_output_map maps logical solid outputs (solid_name, output_name) to particular
# step outputs. This covers the case where a solid maps to multiple steps
# and one wants to be able to attach to the logical output of a solid during execution
class StepBuilderState:
    def __init__(self, pipeline_name, initial_tags=None):
        self.steps = []
        self.step_output_map = StepOutputMap()
        self._tags = check.opt_dict_param(
            initial_tags, 'initial_tags', key_type=str, value_type=str
        )
        self._tags['pipeline'] = pipeline_name

    def get_tags(self, **additional_tags):
        additional_tags.update(self._tags)
        return additional_tags

    @contextmanager
    def push_tags(self, **kwargs):
        old_tags = copy.copy(self._tags)
        self._tags.update(kwargs)
        try:
            yield
        finally:
            self._tags = old_tags



from __future__ import unicode_literals

import subprocess

from dagster.version import __version__


def test_version():
    assert subprocess.check_output(['dagster', '--version']) == bytes(
        ('dagster, version {version}\n'.format(version=__version__)).encode('utf-8')
    )



from abc import ABCMeta, abstractmethod
import pickle

import six

from dagster import check


@six.add_metaclass(ABCMeta)
class SerializationStrategy:
    @abstractmethod
    def serialize_value(self, value, write_file_obj):
        pass

    @abstractmethod
    def deserialize_value(self, read_file_obj):
        pass


class PickleSerializationStrategy(SerializationStrategy):
    def serialize_value(self, value, write_file_obj):
        pickle.dump(value, write_file_obj)

    def deserialize_value(self, read_file_obj):
        return pickle.load(read_file_obj)


@six.add_metaclass(ABCMeta)
class PersistenceStrategy:
    @abstractmethod
    def write_value(self, serialization_strategy, key, value):
        pass

    @abstractmethod
    def read_value(self, serialization_strategy, key):
        pass


class FilePersistencePolicy(PersistenceStrategy):
    def write_value(self, serialization_strategy, key, value):
        check.inst_param(serialization_strategy, 'serialization_strategy', SerializationStrategy)
        check.str_param(key, 'key')

        return serialize_to_file(serialization_strategy, value, key)

    def read_value(self, serialization_strategy, key):
        check.inst_param(serialization_strategy, 'serialization_strategy', SerializationStrategy)
        check.str_param(key, 'key')
        return deserialize_from_file(serialization_strategy, key)


def serialize_to_file(serialization_strategy, value, write_path):
    check.inst_param(serialization_strategy, 'serialization_strategy', SerializationStrategy)
    check.str_param(write_path, 'write_path')

    with open(write_path, 'wb') as write_obj:
        return serialization_strategy.serialize_value(value, write_obj)


def deserialize_from_file(serialization_strategy, read_path):
    check.inst_param(serialization_strategy, 'serialization_strategy', SerializationStrategy)
    check.str_param(read_path, 'read_path')

    with open(read_path, 'rb') as read_obj:
        return serialization_strategy.deserialize_value(read_obj)



from dagster import check

from dagster.core.definitions import (
    ExpectationDefinition,
    ExpectationExecutionInfo,
    InputDefinition,
    OutputDefinition,
    Result,
    Solid,
)

from dagster.core.errors import DagsterExpectationFailedError

from .objects import (
    ExecutionPlanInfo,
    ExecutionStep,
    ExecutionValueSubplan,
    StepInput,
    StepOutput,
    StepOutputHandle,
    StepKind,
    StepBuilderState,
)

from .utility import create_joining_subplan

EXPECTATION_INPUT = 'expectation_input'
EXPECTATION_VALUE_OUTPUT = 'expectation_value'


def _create_expectation_lambda(solid, inout_def, expectation_def, internal_output_name):
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(expectation_def, 'expectations_def', ExpectationDefinition)
    check.str_param(internal_output_name, 'internal_output_name')

    def _do_expectation(context, step, inputs):
        value = inputs[EXPECTATION_INPUT]
        info = ExpectationExecutionInfo(context, inout_def, solid, expectation_def)
        expt_result = expectation_def.expectation_fn(info, value)
        if expt_result.success:
            context.debug(
                'Expectation {key} succeeded on {value}.'.format(key=step.key, value=value)
            )
            yield Result(output_name=internal_output_name, value=inputs[EXPECTATION_INPUT])
        else:
            context.debug('Expectation {key} failed on {value}.'.format(key=step.key, value=value))
            raise DagsterExpectationFailedError(info, value)

    return _do_expectation


def create_expectations_subplan(state, solid, inout_def, prev_step_output_handle, kind):
    check.inst_param(state, 'state', StepBuilderState)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(prev_step_output_handle, 'prev_step_output_handle', StepOutputHandle)
    check.inst_param(kind, 'kind', StepKind)

    input_expect_steps = []
    for expectation_def in inout_def.expectations:
        with state.push_tags(expectation=expectation_def.name):
            expect_step = create_expectation_step(
                state=state,
                solid=solid,
                expectation_def=expectation_def,
                key='{solid}.{desc_key}.{inout_name}.expectation.{expectation_name}'.format(
                    solid=solid.name,
                    desc_key=inout_def.descriptive_key,
                    inout_name=inout_def.name,
                    expectation_name=expectation_def.name,
                ),
                kind=kind,
                prev_step_output_handle=prev_step_output_handle,
                inout_def=inout_def,
            )
            input_expect_steps.append(expect_step)

    return create_joining_subplan(
        state,
        solid,
        '{solid}.{desc_key}.{inout_name}.expectations.join'.format(
            solid=solid.name, desc_key=inout_def.descriptive_key, inout_name=inout_def.name
        ),
        input_expect_steps,
        EXPECTATION_VALUE_OUTPUT,
    )


def create_expectation_step(
    state, solid, expectation_def, key, kind, prev_step_output_handle, inout_def
):
    check.inst_param(state, 'state', StepBuilderState)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(expectation_def, 'input_expct_def', ExpectationDefinition)
    check.inst_param(prev_step_output_handle, 'prev_step_output_handle', StepOutputHandle)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))

    value_type = inout_def.runtime_type

    return ExecutionStep(
        key=key,
        step_inputs=[
            StepInput(
                name=EXPECTATION_INPUT,
                runtime_type=value_type,
                prev_output_handle=prev_step_output_handle,
            )
        ],
        step_outputs=[StepOutput(name=EXPECTATION_VALUE_OUTPUT, runtime_type=value_type)],
        compute_fn=_create_expectation_lambda(
            solid, inout_def, expectation_def, EXPECTATION_VALUE_OUTPUT
        ),
        kind=kind,
        solid=solid,
        tags=state.get_tags(),
    )


def decorate_with_expectations(execution_info, state, solid, transform_step, output_def):
    check.inst_param(execution_info, 'execution_info', ExecutionPlanInfo)
    check.inst_param(state, 'state', StepBuilderState)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(transform_step, 'transform_step', ExecutionStep)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    if execution_info.environment.expectations.evaluate and output_def.expectations:
        return create_expectations_subplan(
            state,
            solid,
            output_def,
            StepOutputHandle(transform_step, output_def.name),
            kind=StepKind.OUTPUT_EXPECTATION,
        )
    else:
        return ExecutionValueSubplan.empty(StepOutputHandle(transform_step, output_def.name))



import itertools
import json
import logging
import uuid

from collections import namedtuple

from dagster import check
from dagster.utils import merge_dicts
from dagster.utils.logging import CompositeLogger, INFO, define_colored_console_logger

from .events import ExecutionEvents
from .types.marshal import PersistenceStrategy


def _kv_message(all_items):
    return ' '.join(
        ['{key}={value}'.format(key=key, value=json.dumps(value)) for key, value in all_items]
    )


DAGSTER_META_KEY = 'dagster_meta'


class ExecutionContext(namedtuple('_ExecutionContext', 'loggers resources tags')):
    def __new__(cls, loggers=None, resources=None, tags=None):
        return super(ExecutionContext, cls).__new__(
            cls,
            loggers=check.opt_list_param(loggers, 'loggers', logging.Logger),
            resources=resources,
            tags=check.opt_dict_param(tags, 'tags'),
        )

    @staticmethod
    def console_logging(log_level=INFO, resources=None):
        return ExecutionContext(
            loggers=[define_colored_console_logger('dagster', log_level)], resources=resources
        )


class RuntimeExecutionContext:
    '''
    A context object flowed through the entire scope of single execution of a
    pipeline of solids. This is used by both framework and user code to log
    messages and metrics. It also maintains a stack of context values so that
    logs, metrics, and any future reporting are reported with a minimal, consistent
    level of context so that developers do not have to repeatedly log well-known
    information (e.g. the name of the solid, the name of the pipeline, etc) when
    logging. Additionally tool author may add their own context values to assist
    reporting.


    Args:
        loggers (List[logging.Logger]):
            The list of loggers that will be invoked whenever the context logging
            functions are called.

        resources(Any):
            An arbitrary user-defined object that can be passed in by a user and
            then access during pipeline execution. This exists so that a user can
            inject their own objects into the context without having to subclass
            ExecutionContext.
    '''

    def __init__(
        self,
        run_id,
        loggers=None,
        resources=None,
        event_callback=None,
        environment_config=None,
        tags=None,
        persistence_policy=None,
    ):

        if loggers is None:
            loggers = [define_colored_console_logger('dagster')]

        self._logger = CompositeLogger(loggers=loggers)
        self.resources = resources
        self._run_id = check.str_param(run_id, 'run_id')
        self._tags = check.opt_dict_param(tags, 'tags')
        self.events = ExecutionEvents(self)

        # For re-construction purposes later on
        self._event_callback = check.opt_callable_param(event_callback, 'event_callback')
        self._environment_config = environment_config
        self.persistence_policy = check.opt_inst_param(
            persistence_policy, 'persistence_policy', PersistenceStrategy
        )

    def for_step(self, step):
        return RuntimeExecutionContext(
            run_id=self.run_id,
            loggers=self._logger.loggers,
            resources=self.resources,
            tags=merge_dicts(step.tags, self._tags),
            environment_config=self.environment_config,
            event_callback=self.event_callback,
            persistence_policy=self.persistence_policy,
        )

    def _log(self, method, orig_message, message_props):
        check.str_param(method, 'method')
        check.str_param(orig_message, 'orig_message')
        check.dict_param(message_props, 'message_props')

        check.invariant(
            'extra' not in message_props, 'do not allow until explicit support is handled'
        )
        check.invariant(
            'exc_info' not in message_props, 'do not allow until explicit support is handled'
        )

        check.invariant('orig_message' not in message_props, 'orig_message reserved value')
        check.invariant('message' not in message_props, 'message reserved value')
        check.invariant('log_message_id' not in message_props, 'log_message_id reserved value')

        log_message_id = str(uuid.uuid4())

        synth_props = {
            'orig_message': orig_message,
            'log_message_id': log_message_id,
            'run_id': self._run_id,
        }

        # We first generate all props for the purpose of producing the semi-structured
        # log message via _kv_messsage
        all_props = dict(
            itertools.chain(synth_props.items(), self._tags.items(), message_props.items())
        )

        message_with_structured_props = _kv_message(all_props.items())

        # So here we use the arbitrary key DAGSTER_META_KEY to store a dictionary of
        # all the meta information that dagster injects into log message.
        # The python logging module, in its infinite wisdom, actually takes all the
        # keys in extra and unconditionally smashes them into the internal dictionary
        # of the logging.LogRecord class. We used a reserved key here to avoid naming
        # collisions with internal variables of the LogRecord class.
        # See __init__.py:363 (makeLogRecord) in the python 3.6 logging module source
        # for the gory details.
        getattr(self._logger, method)(
            message_with_structured_props, extra={DAGSTER_META_KEY: all_props}
        )

    def debug(self, msg, **kwargs):
        '''
        Debug level logging directive. Ends up invoking loggers with DEBUG error level.

        The message will be automatically adorned with context information about the name
        of the pipeline, the name of the solid, and so forth. The use can also add
        context values during execution using the value() method of ExecutionContext.
        Therefore it is generally unnecessary to include this type of information
        (solid name, pipeline name, etc) in the log message unless it is critical
        for the readability/fluency of the log message text itself.

        You can optionally additional context key-value pairs to an individual log
        message using the keyword args to this message

        Args:
            msg (str): The core string
            **kwargs (Dict[str, Any]): Additional context values for only this log message.
        '''
        return self._log('debug', msg, kwargs)

    def info(self, msg, **kwargs):
        '''Log at INFO level

        See debug()'''
        return self._log('info', msg, kwargs)

    def warning(self, msg, **kwargs):
        '''Log at WARNING level

        See debug()'''
        return self._log('warning', msg, kwargs)

    def error(self, msg, **kwargs):
        '''Log at ERROR level

        See debug()'''
        return self._log('error', msg, kwargs)

    def critical(self, msg, **kwargs):
        '''Log at CRITICAL level

        See debug()'''
        return self._log('critical', msg, kwargs)

    def has_tag(self, key):
        check.str_param(key, 'key')
        return key in self._tags

    def get_tag(self, key):
        check.str_param(key, 'key')
        return self._tags[key]

    @property
    def run_id(self):
        return self._run_id

    @property
    def event_callback(self):
        return self._event_callback

    @property
    def has_event_callback(self):
        return self._event_callback is not None

    @property
    def environment_config(self):
        return self._environment_config


class ExecutionMetadata(namedtuple('_ExecutionMetadata', 'run_id tags event_callback loggers')):
    def __new__(cls, run_id=None, tags=None, event_callback=None, loggers=None):
        return super(ExecutionMetadata, cls).__new__(
            cls,
            run_id=check.str_param(run_id, 'run_id') if run_id else str(uuid.uuid4()),
            tags=check.opt_dict_param(tags, 'tags', key_type=str, value_type=str),
            event_callback=check.opt_callable_param(event_callback, 'event_callback'),
            loggers=check.opt_list_param(loggers, 'loggers'),
        )



from dagster import execute_pipeline

from dagster.tutorials.intro_tutorial.reusable_solids import define_reusable_solids_pipeline


def test_run_whole_pipeline():
    pipeline = define_reusable_solids_pipeline()
    pipeline_result = execute_pipeline(
        pipeline,
        {
            'solids': {
                'a_plus_b': {'inputs': {'num1': {'value': 2}, 'num2': {'value': 6}}},
                'c_plus_d': {'inputs': {'num1': {'value': 4}, 'num2': {'value': 8}}},
            }
        },
    )

    assert pipeline_result.success
    assert pipeline_result.result_for_solid('a_plus_b').transformed_value() == 8

    assert pipeline_result.result_for_solid('c_plus_d').transformed_value() == 12
    assert pipeline_result.result_for_solid('final').transformed_value() == 8 * 12



from dagster import execute_pipeline
from dagster.tutorials.intro_tutorial.actual_dag import define_diamond_dag_pipeline
from dagster.tutorials.utils import check_cli_execute_file_pipeline
from dagster.utils import script_relative_path


def test_intro_tutorial_actual_dag():
    pipeline = define_diamond_dag_pipeline()

    result = execute_pipeline(pipeline)

    assert result.success
    assert len(result.result_list) == 4
    assert result.result_for_solid('solid_a').transformed_value() == 1
    assert result.result_for_solid('solid_b').transformed_value() == 2
    assert result.result_for_solid('solid_c').transformed_value() == 3
    assert result.result_for_solid('solid_d').transformed_value() == 6
    return result


def test_intro_tutorial_cli_actual_dag():
    check_cli_execute_file_pipeline(
        script_relative_path('../../../dagster/tutorials/intro_tutorial/actual_dag.py'),
        'define_diamond_dag_pipeline',
    )



from dagster.tutorials.intro_tutorial.tutorial_repository import define_repository


def test_intro_tutorial_repository():
    repo = define_repository()
    assert repo
    assert repo.get_all_pipelines()



from dagster import check
from dagster.core.definitions import Result, Solid, TransformExecutionInfo
from dagster.core.errors import DagsterInvariantViolationError
from dagster.core.execution_context import RuntimeExecutionContext

from .objects import (
    ExecutionPlanInfo,
    ExecutionStep,
    StepBuilderState,
    StepInput,
    StepKind,
    StepOutput,
)


def create_transform_step(execution_info, state, solid, step_inputs, conf):
    check.inst_param(execution_info, 'execution_info', ExecutionPlanInfo)
    check.inst_param(state, 'state', StepBuilderState)
    check.inst_param(solid, 'solid', Solid)
    check.list_param(step_inputs, 'step_inputs', of_type=StepInput)

    return ExecutionStep(
        key='{solid.name}.transform'.format(solid=solid),
        step_inputs=step_inputs,
        step_outputs=[
            StepOutput(name=output_def.name, runtime_type=output_def.runtime_type)
            for output_def in solid.definition.output_defs
        ],
        compute_fn=lambda context, step, inputs: _execute_core_transform(
            execution_info, context, step, conf, inputs
        ),
        kind=StepKind.TRANSFORM,
        solid=solid,
        tags=state.get_tags(),
    )


def _yield_transform_results(execution_info, context, step, conf, inputs):
    gen = step.solid.definition.transform_fn(
        TransformExecutionInfo(context, conf, step.solid, execution_info.pipeline), inputs
    )

    if isinstance(gen, Result):
        raise DagsterInvariantViolationError(
            (
                'Transform for solid {solid_name} returned a Result rather than '
                + 'yielding it. The transform_fn of the core SolidDefinition must yield '
                + 'its results'
            ).format(solid_name=step.solid.name)
        )

    if gen is None:
        return

    for result in gen:
        if not isinstance(result, Result):
            raise DagsterInvariantViolationError(
                (
                    'Transform for solid {solid_name} yielded {result} rather an '
                    + 'an instance of the Result class.'
                ).format(result=repr(result), solid_name=step.solid.name)
            )

        context.info(
            'Solid {solid} emitted output "{output}" value {value}'.format(
                solid=step.solid.name, output=result.output_name, value=repr(result.value)
            )
        )
        yield result


def _execute_core_transform(execution_info, context, step, conf, inputs):
    '''
    Execute the user-specified transform for the solid. Wrap in an error boundary and do
    all relevant logging and metrics tracking
    '''
    check.inst_param(execution_info, 'execution_info', ExecutionPlanInfo)
    check.inst_param(context, 'context', RuntimeExecutionContext)
    check.inst_param(step, 'step', ExecutionStep)
    check.dict_param(inputs, 'inputs', key_type=str)

    solid = step.solid

    context.debug('Executing core transform for solid {solid}.'.format(solid=solid.name))

    all_results = list(_yield_transform_results(execution_info, context, step, conf, inputs))

    if len(all_results) != len(solid.definition.output_defs):
        emitted_result_names = {r.output_name for r in all_results}
        solid_output_names = {output_def.name for output_def in solid.definition.output_defs}
        omitted_outputs = solid_output_names.difference(emitted_result_names)
        context.info(
            'Solid {solid} did not fire outputs {outputs}'.format(
                solid=solid.name, outputs=repr(omitted_outputs)
            )
        )

    return all_results



from collections import namedtuple
from enum import Enum
import imp
import sys
import importlib
import os

import click

from dagster import PipelineDefinition, RepositoryDefinition, check

from dagster.utils import load_yaml_from_path

if sys.version_info[0] >= 3:
    import reloader

    # The reloader module allows us to specify a blacklist of modules not to reload,
    # but we want something a bit more flexible. We only want to reload the user's
    # pipeline code, not dagster imports, installed packages, or parts of the python
    # install. Some of these things (like numpy) actually cannot be reloaded.
    #
    # Note: We cannot hot-reload Dagster itself. Re-loading RepositoryDefinition, etc.
    # in user code causes `instanceof` checks to fail since this code still references
    # the original copies of RespositoryDefinition, etc.
    #

    # allow use of protected member
    # pylint: disable=W0212
    _reload = reloader._reload

    def conditional_reload(m, visited):
        if "/usr/local" in m.__file__ or "site-packages" in m.__file__:
            return
        if m.__name__.startswith("dagster.core") or m.__name__.startswith("dagster.utils"):
            return
        _reload(m, visited)

    reloader._reload = conditional_reload
else:

    class ReloaderStub:
        def reload(self, module):
            pass

        def enable(self):
            sys.stderr.write('Hot-reloading only supports Python 3+')

    reloader = ReloaderStub()

INFO_FIELDS = set(['repository_yaml', 'pipeline_name', 'python_file', 'fn_name', 'module_name'])

PipelineTargetInfo = namedtuple(
    'PipelineTargetInfo', 'repository_yaml pipeline_name python_file fn_name module_name'
)


class RepositoryTargetMode(Enum):
    YAML_FILE = 0
    MODULE = 1
    FILE = 2


RepositoryTargetInfo = namedtuple(
    'RepositoryTargetInfo', 'repository_yaml module_name python_file fn_name'
)


class PipelineTargetMode(Enum):
    PIPELINE_PYTHON_FILE = 0
    PIPELINE_MODULE = 1
    REPOSITORY_PYTHON_FILE = 2
    REPOSITORY_MODULE = 3
    REPOSITORY_YAML_FILE = 4


FileTargetFunction = namedtuple('FileTargetFunction', 'python_file fn_name')

ModuleTargetFunction = namedtuple('ModuleTargetFunction', 'module_name fn_name')

RepositoryPythonFileData = namedtuple(
    'RepositoryPythonFileData', 'file_target_function pipeline_name'
)

RepositoryModuleData = namedtuple('RepositoryModuleData', 'module_target_function pipeline_name')

RepositoryYamlData = namedtuple('RepositoryYamlData', 'repository_yaml pipeline_name')

PipelineLoadingModeData = namedtuple('PipelineLoadingModeData', 'mode data')

RepositoryLoadingModeData = namedtuple('RepositoryLoadingModeData', 'mode data')


class InvalidPipelineLoadingComboError(Exception):
    pass


class InvalidRepositoryLoadingComboError(Exception):
    pass


def check_info_fields(info, *fields):
    check.inst_param(info, 'info', PipelineTargetInfo)
    check.tuple_param(fields, 'fields')

    info_dict = info._asdict()
    for field in fields:
        if info_dict[field] is None:
            return False

    for none_field in INFO_FIELDS.difference(set(fields)):
        if info_dict[none_field] is not None:
            raise InvalidPipelineLoadingComboError(
                (
                    'field: {none_field} with value {value} should not be set if'
                    '{fields} were provided'
                ).format(
                    value=repr(info_dict[none_field]), none_field=none_field, fields=repr(fields)
                )
            )

    return True


def repo_load_invariant(condition):
    if not condition:
        raise InvalidRepositoryLoadingComboError()


def create_repository_loading_mode_data(info):
    check.inst_param(info, 'info', RepositoryTargetInfo)

    if info.repository_yaml:
        repo_load_invariant(info.module_name is None)
        repo_load_invariant(info.python_file is None)
        repo_load_invariant(info.fn_name is None)
        return RepositoryLoadingModeData(
            mode=RepositoryTargetMode.YAML_FILE, data=info.repository_yaml
        )
    elif info.module_name and info.fn_name:
        repo_load_invariant(info.repository_yaml is None)
        repo_load_invariant(info.python_file is None)
        return RepositoryLoadingModeData(
            mode=RepositoryTargetMode.MODULE,
            data=ModuleTargetFunction(module_name=info.module_name, fn_name=info.fn_name),
        )
    elif info.python_file and info.fn_name:
        repo_load_invariant(info.repository_yaml is None)
        repo_load_invariant(info.module_name is None)
        return RepositoryLoadingModeData(
            mode=RepositoryTargetMode.FILE,
            data=FileTargetFunction(python_file=info.python_file, fn_name=info.fn_name),
        )
    else:
        raise InvalidRepositoryLoadingComboError()


def create_pipeline_loading_mode_data(info):
    check.inst_param(info, 'info', PipelineTargetInfo)

    if check_info_fields(info, 'python_file', 'fn_name', 'pipeline_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.REPOSITORY_PYTHON_FILE,
            data=RepositoryPythonFileData(
                file_target_function=FileTargetFunction(
                    python_file=info.python_file, fn_name=info.fn_name
                ),
                pipeline_name=info.pipeline_name,
            ),
        )
    elif check_info_fields(info, 'module_name', 'fn_name', 'pipeline_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.REPOSITORY_MODULE,
            data=RepositoryModuleData(
                module_target_function=ModuleTargetFunction(
                    module_name=info.module_name, fn_name=info.fn_name
                ),
                pipeline_name=info.pipeline_name,
            ),
        )
    elif check_info_fields(info, 'python_file', 'fn_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.PIPELINE_PYTHON_FILE,
            data=FileTargetFunction(python_file=info.python_file, fn_name=info.fn_name),
        )
    elif check_info_fields(info, 'module_name', 'fn_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.PIPELINE_MODULE,
            data=ModuleTargetFunction(module_name=info.module_name, fn_name=info.fn_name),
        )
    elif info.pipeline_name:
        for none_field in ['python_file', 'fn_name', 'module_name']:
            if getattr(info, none_field) is not None:
                raise InvalidPipelineLoadingComboError(
                    '{none_field} is not None. Got {value}'.format(
                        none_field=none_field, value=getattr(info, none_field)
                    )
                )

        check.invariant(info.repository_yaml is not None)

        return PipelineLoadingModeData(
            mode=PipelineTargetMode.REPOSITORY_YAML_FILE,
            data=RepositoryYamlData(
                repository_yaml=info.repository_yaml, pipeline_name=info.pipeline_name
            ),
        )
    else:
        raise InvalidPipelineLoadingComboError()


class DynamicObject:
    def __init__(self, module, module_name, fn_name):
        self.module = module
        self.module_name = module_name
        self.fn_name = fn_name
        self.object = None
        self.coerce_to_repo = False
        self.loaded = False

    def load(self):
        if self.loaded:
            reloader.reload(self.module)
        self.loaded = True

        fn = getattr(self.module, self.fn_name)
        check.is_callable(fn)
        obj = fn()

        # Eventually this class will be generic and not coupled to
        # Pipeline / Repository types. Tracking this issue here:
        # https://github.com/dagster-io/dagster/issues/246
        if self.coerce_to_repo:
            if isinstance(obj, RepositoryDefinition):
                self.object = obj
            elif isinstance(obj, PipelineDefinition):
                self.object = RepositoryDefinition(
                    name=EMPHERMAL_NAME, pipeline_dict={obj.name: lambda: obj}
                )
            else:
                raise InvalidPipelineLoadingComboError(
                    'entry point must return a repository or pipeline'
                )
        else:
            self.object = obj

        return self.object


def load_file_target_function(file_target_function):
    reloader.enable()
    check.inst_param(file_target_function, 'file_target_function', FileTargetFunction)
    module_name = os.path.splitext(os.path.basename(file_target_function.python_file))[0]
    module = imp.load_source(module_name, file_target_function.python_file)
    return DynamicObject(module, module_name, file_target_function.fn_name)


def load_module_target_function(module_target_function):
    reloader.enable()
    check.inst_param(module_target_function, 'module_target_function', ModuleTargetFunction)
    module = importlib.import_module(module_target_function.module_name)
    return DynamicObject(module, module_target_function.module_name, module_target_function.fn_name)


EMPHERMAL_NAME = '<<unnamed>>'


def load_repository_object_from_target_info(info):
    check.inst_param(info, 'info', RepositoryTargetInfo)

    mode_data = create_repository_loading_mode_data(info)

    if mode_data.mode == RepositoryTargetMode.YAML_FILE:
        dynamic_obj = load_repository_from_file(mode_data.data)
    elif mode_data.mode == RepositoryTargetMode.MODULE:
        dynamic_obj = load_module_target_function(mode_data.data)
    elif mode_data.mode == RepositoryTargetMode.FILE:
        dynamic_obj = load_file_target_function(mode_data.data)
    else:
        check.failed('should not reach')

    dynamic_obj.coerce_to_repo = True
    return dynamic_obj


def load_repository_from_target_info(info):
    return check.inst(load_repository_object_from_target_info(info).load(), RepositoryDefinition)


# Keeping this code around for a week. I might need to be able to
# coerce a single pipeline repo into a pipeline at some point while
# we work out the kinks in the command line tool.
#
# If this is still around in a week or two delete this -- schrockn (09/18/18)

# def _pipeline_from_dynamic_object(mode_data, dynamic_object):
#     check.inst_param(mode_data, 'mode_data', PipelineLoadingModeData)
#     check.inst_param(dynamic_object, 'dynamic_object', DynamicObject)

#     repository = check.inst(
#         ensure_in_repo(dynamic_object).object,
#         RepositoryDefinition,
#     )
#     if len(repository.pipeline_dict) == 1:
#         return repository.get_all_pipelines()[0]

#     return repository.get_pipeline(mode_data.data.pipeline_name)


def load_pipeline_from_target_info(info):
    check.inst_param(info, 'info', PipelineTargetInfo)

    mode_data = create_pipeline_loading_mode_data(info)

    if mode_data.mode == PipelineTargetMode.REPOSITORY_PYTHON_FILE:
        repository = check.inst(
            load_file_target_function(mode_data.data.file_target_function).load(),
            RepositoryDefinition,
        )
        return repository.get_pipeline(mode_data.data.pipeline_name)
        # dynamic_object = load_file_target_function(mode_data.data.file_target_function)
        # return _pipeline_from_dynamic_object(mode_data, dynamic_object)
        # If this is still around in a week or two delete this -- schrockn (09/18/18)
    elif mode_data.mode == PipelineTargetMode.REPOSITORY_MODULE:
        repository = check.inst(
            load_module_target_function(mode_data.data.module_target_function).load(),
            RepositoryDefinition,
        )
        return repository.get_pipeline(mode_data.data.pipeline_name)
        # dynamic_object = load_module_target_function(mode_data.data.module_target_function)
        # return _pipeline_from_dynamic_object(mode_data, dynamic_object)
        # If this is still around in a week or two delete this -- schrockn (09/18/18)
    elif mode_data.mode == PipelineTargetMode.PIPELINE_PYTHON_FILE:
        return check.inst(load_file_target_function(mode_data.data).load(), PipelineDefinition)
    elif mode_data.mode == PipelineTargetMode.PIPELINE_MODULE:
        return check.inst(load_module_target_function(mode_data.data).load(), PipelineDefinition)
    elif mode_data.mode == PipelineTargetMode.REPOSITORY_YAML_FILE:
        repository = check.inst(
            load_repository_from_file(mode_data.data.repository_yaml).load(), RepositoryDefinition
        )
        return repository.get_pipeline(mode_data.data.pipeline_name)
    else:
        check.failed('Should never reach')


def load_repository_from_file(file_path):
    check.str_param(file_path, 'file_path')

    config = load_yaml_from_path(file_path)
    repository_config = check.dict_elem(config, 'repository')
    module_name = check.opt_str_elem(repository_config, 'module')
    file_name = check.opt_str_elem(repository_config, 'file')
    fn_name = check.str_elem(repository_config, 'fn')

    if module_name:
        return load_module_target_function(ModuleTargetFunction(module_name, fn_name))
    else:
        # rebase file in config off of the path in the config file
        file_name = os.path.join(os.path.dirname(os.path.abspath(file_path)), file_name)
        return load_file_target_function(FileTargetFunction(file_name, fn_name))


def apply_click_params(command, *click_params):
    for click_param in click_params:
        command = click_param(command)
    return command


def repository_target_argument(f):
    return apply_click_params(
        f,
        click.option(
            '--repository-yaml',
            '-y',
            type=click.STRING,
            help=(
                'Path to config file. Defaults to ./repository.yml. if --python-file '
                'and --module-name are not specified'
            ),
        ),
        click.option(
            '--python-file',
            '-f',
            help='Specify python file where repository or pipeline function lives.',
        ),
        click.option(
            '--module-name', '-m', help='Specify module where repository or pipeline function lives'
        ),
        click.option('--fn-name', '-n', help='Function that returns either repository or pipeline'),
    )


def pipeline_target_command(f):
    # f = repository_config_argument(f)
    # nargs=-1 is used right now to make this argument optional
    # it can only handle 0 or 1 pipeline names
    # see create_pipeline_from_cli_args
    return apply_click_params(
        f,
        click.option(
            '--repository-yaml',
            '-y',
            type=click.STRING,
            help=(
                'Path to config file. Defaults to ./repository.yml. if --python-file '
                'and --module-name are not specified'
            ),
        ),
        click.argument('pipeline_name', nargs=-1),
        click.option('--python-file', '-f'),
        click.option('--module-name', '-m'),
        click.option('--fn-name', '-n'),
    )


def all_none(kwargs):
    for value in kwargs.values():
        if value is not None:
            return False
    return True


def load_target_info_from_cli_args(cli_args):
    check.dict_param(cli_args, 'cli_args')

    if all_none(cli_args):
        cli_args['repository_yaml'] = 'repository.yml'

    return RepositoryTargetInfo(**cli_args)



from dagster import check
from .config import ConfigType, ConfigTypeAttributes, DEFAULT_TYPE_ATTRIBUTES
from .default_applier import apply_default_values


def all_optional_type(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)

    if config_type.is_composite or config_type.is_selector:
        for field in config_type.fields.values():
            if not field.is_optional:
                return False
        return True

    return False


class __FieldValueSentinel:
    pass


class __InferOptionalCompositeFieldSentinel:
    pass


FIELD_NO_DEFAULT_PROVIDED = __FieldValueSentinel

INFER_OPTIONAL_COMPOSITE_FIELD = __InferOptionalCompositeFieldSentinel


def check_field_param(obj, param_name):
    return check.inst_param(obj, param_name, FieldImpl)


def check_opt_field_param(obj, param_name):
    return check.opt_inst_param(obj, param_name, FieldImpl)


class FieldImpl:
    '''
    A field in a config object.

    Attributes:
        config_type (ConfigType): The type of the field.
        default_value (Any):
            If the Field is optional, a default value can be provided when the field value
            is not specified.
        is_optional (bool): Is the field optional.
        description (str): Description of the field.
    '''

    def __init__(
        self,
        config_type,
        default_value=FIELD_NO_DEFAULT_PROVIDED,
        is_optional=INFER_OPTIONAL_COMPOSITE_FIELD,
        description=None,
    ):
        self.config_type = check.inst_param(config_type, 'config_type', ConfigType)

        self.description = check.opt_str_param(description, 'description')
        if is_optional == INFER_OPTIONAL_COMPOSITE_FIELD:
            is_optional = all_optional_type(self.config_type)
            if is_optional is True:
                self._default_value = apply_default_values(self.config_type, None)
            else:
                self._default_value = default_value
        else:
            is_optional = check.bool_param(is_optional, 'is_optional')
            self._default_value = default_value

        if is_optional is False:
            check.param_invariant(
                default_value == FIELD_NO_DEFAULT_PROVIDED,
                'default_value',
                'required arguments should not specify default values',
            )

        self.is_optional = is_optional

    @property
    def is_required(self):
        return not self.is_optional

    @property
    def default_provided(self):
        '''Was a default value provided

        Returns:
            bool: Yes or no
        '''
        return self._default_value != FIELD_NO_DEFAULT_PROVIDED

    @property
    def default_value(self):
        check.invariant(self.default_provided, 'Asking for default value when none was provided')

        if callable(self._default_value):
            return self._default_value()

        return self._default_value

    @property
    def default_value_as_str(self):
        check.invariant(self.default_provided, 'Asking for default value when none was provided')

        if callable(self._default_value):
            return repr(self._default_value)

        return str(self._default_value)


class _ConfigHasFields(ConfigType):
    def __init__(self, fields, *args, **kwargs):
        self.fields = check.dict_param(fields, 'fields', key_type=str, value_type=FieldImpl)
        super(_ConfigHasFields, self).__init__(*args, **kwargs)

    @property
    def inner_types(self):
        return list(set(self._yield_inner_types()))

    def _yield_inner_types(self):
        for field in self.fields.values():
            yield field.config_type
            for inner_type in field.config_type.inner_types:
                yield inner_type


class _ConfigComposite(_ConfigHasFields):
    @property
    def is_composite(self):
        return True


class _ConfigSelector(_ConfigHasFields):
    @property
    def is_selector(self):
        return True


# HACK HACK HACK
#
# This is not good and a better solution needs to be found. In order
# for the client-side typeahead in dagit to work as currently structured,
# dictionaries need names. While we deal with that we're going to automatically
# name dictionaries. This will cause odd behavior and bugs is you restart
# the server-side process, the type names changes, and you do not refresh the client.
#
# A possible short term mitigation would to name the dictionary based on the hash
# of its member fields to provide stability in between process restarts.
#
class DictCounter:
    _count = 0

    @staticmethod
    def get_next_count():
        DictCounter._count += 1
        return DictCounter._count


def NamedDict(name, fields, description=None, type_attributes=DEFAULT_TYPE_ATTRIBUTES):
    class _NamedDict(_ConfigComposite):
        def __init__(self):
            super(_NamedDict, self).__init__(
                key=name,
                name=name,
                fields=fields,
                description=description,
                type_attributes=type_attributes,
            )

    return _NamedDict


def Dict(fields):
    class _Dict(_ConfigComposite):
        def __init__(self):
            key = 'Dict.' + str(DictCounter.get_next_count())
            super(_Dict, self).__init__(
                name=None,
                key=key,
                fields=fields,
                description='A configuration dictionary with typed fields',
                type_attributes=ConfigTypeAttributes(is_builtin=True),
            )

    return _Dict


def Selector(fields):
    class _Selector(_ConfigSelector):
        def __init__(self):
            key = 'Selector.' + str(DictCounter.get_next_count())
            super(_Selector, self).__init__(
                key=key,
                name=None,
                fields=fields,
                # description='A configuration dictionary with typed fields',
                type_attributes=ConfigTypeAttributes(is_builtin=True),
            )

    return _Selector


def NamedSelector(name, fields, description=None, type_attributes=DEFAULT_TYPE_ATTRIBUTES):
    check.str_param(name, 'name')

    class _NamedSelector(_ConfigSelector):
        def __init__(self):
            super(_NamedSelector, self).__init__(
                key=name,
                name=name,
                fields=fields,
                description=description,
                type_attributes=type_attributes,
            )

    return _NamedSelector



from __future__ import print_function

from dagster.cli.pipeline import execute_execute_command
from dagster.utils import pushd, script_relative_path


def test_execute_pipeline():
    execute_kwargs = {
        'pipeline_name': ['demo_pipeline'],
        'repository_yaml': 'pipeline_execution_repository.yml',
        'module_name': None,
        'python_file': None,
        'fn_name': None,
    }
    with pushd(script_relative_path('../../../dagster/tutorials/intro_tutorial/')):
        execute_execute_command(['pipeline_execution_env.yml'], execute_kwargs, print)



from dagster import check

from .builtin_enum import BuiltinEnum
from .wrapping import WrappingType

MAGIC_RUNTIME_TYPE_NAME = '__runtime_type'


def is_runtime_type_decorated_klass(klass):
    check.type_param(klass, 'klass')
    return hasattr(klass, MAGIC_RUNTIME_TYPE_NAME)


def check_dagster_type_param(dagster_type, param_name, base_type):

    # Cannot check base_type because of circular references and no fwd declarations
    if dagster_type is None:
        return dagster_type
    if isinstance(dagster_type, BuiltinEnum):
        return dagster_type
    if isinstance(dagster_type, WrappingType):
        return dagster_type
    if is_runtime_type_decorated_klass(dagster_type):
        return dagster_type

    check.param_invariant(
        isinstance(dagster_type, type),
        'dagster_type',
        'Invalid dagster_type got {dagster_type}'.format(dagster_type=dagster_type),
    )

    if not issubclass(dagster_type, base_type):
        check.failed(
            (
                'Parameter {param_name} must be a valid dagster type: A builtin (e.g. String, Int, '
                'etc), a wrapping type (List or Nullable), or a type class. Got {dagster_type}'
            ).format(param_name=param_name, dagster_type=dagster_type)
        )

    return dagster_type



import pytest

from dagster import DagsterExpectationFailedError, execute_pipeline
from dagster.tutorials.intro_tutorial.expectations import define_expectations_tutorial_pipeline


def test_intro_tutorial_expectations_step_one():
    result = execute_pipeline(
        define_expectations_tutorial_pipeline(),
        {
            'context': {'default': {'config': {'log_level': 'DEBUG'}}},
            'solids': {'add_ints': {'inputs': {'num_one': {'value': 2}, 'num_two': {'value': 3}}}},
        },
    )

    assert result.success


def define_failing_environment_config():
    return {
        'context': {'default': {'config': {'log_level': 'DEBUG'}}},
        'solids': {'add_ints': {'inputs': {'num_one': {'value': -2}, 'num_two': {'value': 3}}}},
    }


def test_intro_tutorial_expectations_step_two_fails_hard():
    with pytest.raises(DagsterExpectationFailedError):
        execute_pipeline(
            define_expectations_tutorial_pipeline(), define_failing_environment_config()
        )


def test_intro_tutorial_expectations_step_two_fails_soft():
    result = execute_pipeline(
        define_expectations_tutorial_pipeline(),
        define_failing_environment_config(),
        throw_on_error=False,
    )

    assert not result.success



from dagster import Field, Dict, String, Int, Bool, Nullable, List, Selector, Any

from dagster.core.types.evaluator import (
    DagsterEvaluationErrorReason,
    EvaluationStackListItemEntry,
    EvaluationStackPathEntry,
    EvaluateValueResult,
    evaluate_config_value,
)
from dagster.core.types.field import resolve_to_config_type


def eval_config_value_from_dagster_type(dagster_type, value):
    return evaluate_config_value(resolve_to_config_type(dagster_type), value)


def assert_success(result, expected_value):
    assert result.success
    assert result.value == expected_value


def test_evaluate_scalar_success():
    assert_success(eval_config_value_from_dagster_type(String, 'foobar'), 'foobar')
    assert_success(eval_config_value_from_dagster_type(Int, 34234), 34234)
    assert_success(eval_config_value_from_dagster_type(Bool, True), True)


def test_evaluate_scalar_failure():
    result = eval_config_value_from_dagster_type(String, 2343)
    assert not result.success
    assert result.value is None
    assert len(result.errors) == 1
    error = result.errors[0]
    assert error.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH
    assert not error.stack.entries
    assert error.error_data.config_type.name == 'String'
    assert error.error_data.value_rep == '2343'


SingleLevelDict = Dict({'level_one': Field(String)})


def test_single_error():
    success_value = {'level_one': 'ksjdfd'}
    assert_success(
        eval_config_value_from_dagster_type(SingleLevelDict, success_value), success_value
    )


def test_single_level_scalar_mismatch():
    value = {'level_one': 234}
    result = eval_config_value_from_dagster_type(SingleLevelDict, value)
    assert not result.success
    assert result.value is None
    assert len(result.errors) == 1
    error = result.errors[0]
    assert error.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH
    assert len(error.stack.entries) == 1
    assert error.stack.entries[0].field_name == 'level_one'
    assert error.stack.entries[0].field_def.config_type.name == 'String'


def test_single_level_dict_not_a_dict():
    value = 'not_a_dict'
    result = eval_config_value_from_dagster_type(SingleLevelDict, value)
    assert not result.success
    assert result.value is None
    assert len(result.errors) == 1
    error = result.errors[0]
    assert error.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH
    assert not error.stack.entries


def test_root_missing_field():
    result = eval_config_value_from_dagster_type(SingleLevelDict, {})
    assert not result.success
    assert result.value is None
    assert len(result.errors) == 1
    error = result.errors[0]
    assert error.reason == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD
    assert len(result.errors_at_level()) == 1
    assert error.error_data.field_name == 'level_one'


DoubleLevelDict = Dict(
    {
        'level_one': Field(
            Dict(
                {
                    'string_field': Field(String),
                    'int_field': Field(Int, is_optional=True, default_value=989),
                    'bool_field': Field(Bool),
                }
            )
        )
    }
)


def test_nested_success():
    value = {'level_one': {'string_field': 'skdsjfkdj', 'int_field': 123, 'bool_field': True}}

    assert_success(eval_config_value_from_dagster_type(DoubleLevelDict, value), value)

    result = eval_config_value_from_dagster_type(
        DoubleLevelDict, {'level_one': {'string_field': 'kjfkd', 'bool_field': True}}
    )

    assert isinstance(result, EvaluateValueResult)

    assert result.success
    assert result.value['level_one']['int_field'] == 989


def test_nested_error_one_field_not_defined():
    value = {
        'level_one': {
            'string_field': 'skdsjfkdj',
            'int_field': 123,
            'bool_field': True,
            'no_field_one': 'kdjfkd',
        }
    }

    result = eval_config_value_from_dagster_type(DoubleLevelDict, value)

    assert not result.success
    assert len(result.errors) == 1
    error = result.errors[0]
    assert error.reason == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED
    assert error.error_data.field_name == 'no_field_one'
    assert len(error.stack.entries) == 1
    stack_entry = error.stack.entries[0]
    assert stack_entry.field_name == 'level_one'
    assert 'Dict' in stack_entry.field_def.config_type.key


def get_field_name_error(result, field_name):
    for error in result.errors:
        if error.error_data.field_name == field_name:
            return error
    assert False


def test_nested_error_two_fields_not_defined():
    value = {
        'level_one': {
            'string_field': 'skdsjfkdj',
            'int_field': 123,
            'bool_field': True,
            'no_field_one': 'kdjfkd',
            'no_field_two': 'kdjfkd',
        }
    }

    result = eval_config_value_from_dagster_type(DoubleLevelDict, value)

    assert not result.success
    assert len(result.errors) == 2

    field_one_error = get_field_name_error(result, 'no_field_one')
    field_two_error = get_field_name_error(result, 'no_field_two')

    assert field_one_error.reason == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED
    assert field_one_error.error_data.field_name == 'no_field_one'
    assert field_two_error.reason == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED
    assert field_two_error.error_data.field_name == 'no_field_two'


def test_nested_error_missing_fields():
    value = {'level_one': {'string_field': 'skdsjfkdj'}}

    result = eval_config_value_from_dagster_type(DoubleLevelDict, value)
    assert not result.success
    assert len(result.errors) == 1
    error = result.errors[0]
    assert error.reason == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD
    assert error.error_data.field_name == 'bool_field'


def test_nested_error_multiple_missing_fields():
    value = {'level_one': {}}

    result = eval_config_value_from_dagster_type(DoubleLevelDict, value)
    assert not result.success
    assert len(result.errors) == 2

    assert (
        get_field_name_error(result, 'bool_field').reason
        == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD
    )
    assert (
        get_field_name_error(result, 'string_field').reason
        == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD
    )


def test_nested_missing_and_not_defined():
    value = {'level_one': {'not_defined': 'kjdfkdj'}}

    result = eval_config_value_from_dagster_type(DoubleLevelDict, value)
    assert not result.success
    assert len(result.errors) == 3

    assert (
        get_field_name_error(result, 'bool_field').reason
        == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD
    )

    assert (
        get_field_name_error(result, 'string_field').reason
        == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD
    )

    assert (
        get_field_name_error(result, 'not_defined').reason
        == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED
    )


MultiLevelDictType = Dict(
    {
        'level_one_string_field': Field(String),
        'level_two_dict': Field(
            Dict(
                {
                    'level_two_int_field': Field(Int),
                    'level_three_dict': Field(Dict({'level_three_string': Field(String)})),
                }
            )
        ),
    }
)


def test_multilevel_success():
    working_value = {
        'level_one_string_field': 'foo',
        'level_two_dict': {
            'level_two_int_field': 234234,
            'level_three_dict': {'level_three_string': 'kjdfkd'},
        },
    }

    assert_success(
        eval_config_value_from_dagster_type(MultiLevelDictType, working_value), working_value
    )


def test_deep_scalar():
    value = {
        'level_one_string_field': 'foo',
        'level_two_dict': {
            'level_two_int_field': 234234,
            'level_three_dict': {'level_three_string': 123},
        },
    }

    result = eval_config_value_from_dagster_type(MultiLevelDictType, value)
    assert not result.success
    assert len(result.errors) == 1
    error = result.errors[0]
    assert error.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH
    assert error.error_data.config_type.name == 'String'
    assert error.error_data.value_rep == '123'
    assert len(error.stack.entries) == 3

    assert [entry.field_name for entry in error.stack.entries] == [
        'level_two_dict',
        'level_three_dict',
        'level_three_string',
    ]

    assert not result.errors_at_level('level_one_string_field')
    assert not result.errors_at_level('level_two_dict')
    assert not result.errors_at_level('level_two_dict', 'level_three_dict')
    assert (
        len(result.errors_at_level('level_two_dict', 'level_three_dict', 'level_three_string')) == 1
    )


def test_deep_mixed_level_errors():
    value = {
        'level_one_string_field': 'foo',
        'level_one_not_defined': 'kjsdkfjd',
        'level_two_dict': {
            # 'level_two_int_field': 234234, # missing
            'level_three_dict': {'level_three_string': 123}
        },
    }

    result = eval_config_value_from_dagster_type(MultiLevelDictType, value)
    assert not result.success
    assert len(result.errors) == 3

    root_errors = result.errors_at_level()
    assert len(root_errors) == 1
    root_error = root_errors[0]
    assert root_error.reason == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED
    assert root_error.error_data.field_name == 'level_one_not_defined'

    level_two_errors = result.errors_at_level('level_two_dict')
    assert len(level_two_errors) == 1
    level_two_error = level_two_errors[0]
    assert level_two_error.reason == DagsterEvaluationErrorReason.MISSING_REQUIRED_FIELD
    assert level_two_error.error_data.field_name == 'level_two_int_field'

    assert not result.errors_at_level('level_two_dict', 'level_three_dict')

    final_level_errors = result.errors_at_level(
        'level_two_dict', 'level_three_dict', 'level_three_string'
    )

    assert len(final_level_errors) == 1
    final_level_error = final_level_errors[0]

    assert final_level_error.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH


ExampleSelector = Selector({'option_one': Field(String), 'option_two': Field(String)})


def test_example_selector_success():
    result = eval_config_value_from_dagster_type(ExampleSelector, {'option_one': 'foo'})
    assert result.success
    assert result.value == {'option_one': 'foo'}

    result = eval_config_value_from_dagster_type(ExampleSelector, {'option_two': 'foo'})
    assert result.success
    assert result.value == {'option_two': 'foo'}


def test_example_selector_error_top_level_type():
    result = eval_config_value_from_dagster_type(ExampleSelector, 'kjsdkf')
    assert not result.success
    assert result.value is None
    assert len(result.errors) == 1
    assert result.errors[0].reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH


def test_example_selector_wrong_field():
    result = eval_config_value_from_dagster_type(ExampleSelector, {'nope': 234})
    assert not result.success
    assert result.value is None
    assert len(result.errors) == 1
    assert result.errors[0].reason == DagsterEvaluationErrorReason.FIELD_NOT_DEFINED


def test_example_selector_multiple_fields():
    result = eval_config_value_from_dagster_type(
        ExampleSelector, {'option_one': 'foo', 'option_two': 'boo'}
    )

    assert not result.success
    assert len(result.errors) == 1
    assert result.errors[0].reason == DagsterEvaluationErrorReason.SELECTOR_FIELD_ERROR


SelectorWithDefaults = Selector({'default': Field(String, is_optional=True, default_value='foo')})


def test_selector_with_defaults():
    result = eval_config_value_from_dagster_type(SelectorWithDefaults, {})
    assert result.success
    assert result.value == {'default': 'foo'}


def test_evaluate_list_string():
    string_list = List(String)
    result = eval_config_value_from_dagster_type(string_list, ["foo"])
    assert result.success
    assert result.value == ["foo"]


def test_evaluate_list_error_item_mismatch():
    string_list = List(String)
    result = eval_config_value_from_dagster_type(string_list, [1])
    assert not result.success
    assert len(result.errors) == 1
    assert result.errors[0].reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH


def test_evaluate_list_error_top_level_mismatch():
    string_list = List(String)
    result = eval_config_value_from_dagster_type(string_list, 1)
    assert not result.success
    assert len(result.errors) == 1
    assert result.errors[0].reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH


def test_evaluate_double_list():
    string_double_list = List(List(String))
    result = eval_config_value_from_dagster_type(string_double_list, [['foo']])
    assert result.success
    assert result.value == [['foo']]


def test_config_list_in_dict():
    nested_list = Dict({'nested_list': Field(List(Int))})

    value = {'nested_list': [1, 2, 3]}
    result = eval_config_value_from_dagster_type(nested_list, value)
    assert result.success
    assert result.value == value


def test_config_list_in_dict_error():
    nested_list = Dict({'nested_list': Field(List(Int))})

    value = {'nested_list': [1, 'bar', 3]}
    result = eval_config_value_from_dagster_type(nested_list, value)
    assert not result.success
    assert len(result.errors) == 1
    error = result.errors[0]
    assert error.reason == DagsterEvaluationErrorReason.RUNTIME_TYPE_MISMATCH
    assert len(error.stack.entries) == 2
    stack_entry = error.stack.entries[0]
    assert isinstance(stack_entry, EvaluationStackPathEntry)
    assert stack_entry.field_name == 'nested_list'
    list_entry = error.stack.entries[1]
    assert isinstance(list_entry, EvaluationStackListItemEntry)
    assert list_entry.list_index == 1


def test_config_double_list():
    nested_lists = Dict(
        {'nested_list_one': Field(List(Int)), 'nested_list_two': Field(List(String))}
    )

    value = {'nested_list_one': [1, 2, 3], 'nested_list_two': ['foo', 'bar']}

    result = eval_config_value_from_dagster_type(nested_lists, value)
    assert result.success
    assert result.value == value

    error_value = {'nested_list_one': 'kjdfkdj', 'nested_list_two': ['bar']}

    error_result = eval_config_value_from_dagster_type(nested_lists, error_value)
    assert not error_result.success


def test_config_double_list_double_error():
    nested_lists = Dict(
        fields={'nested_list_one': Field(List(Int)), 'nested_list_two': Field(List(String))}
    )

    error_value = {'nested_list_one': 'kjdfkdj', 'nested_list_two': ['bar', 2]}
    error_result = eval_config_value_from_dagster_type(nested_lists, error_value)
    assert not error_result.success
    assert len(error_result.errors) == 2


def test_nullable_int():
    assert not eval_config_value_from_dagster_type(Int, None).success
    assert eval_config_value_from_dagster_type(Int, 0).success
    assert eval_config_value_from_dagster_type(Int, 1).success

    assert eval_config_value_from_dagster_type(Nullable(Int), None).success
    assert eval_config_value_from_dagster_type(Nullable(Int), 0).success
    assert eval_config_value_from_dagster_type(Nullable(Int), 1).success


def test_nullable_list():
    list_of_ints = List(Int)

    assert not eval_config_value_from_dagster_type(list_of_ints, None).success
    assert eval_config_value_from_dagster_type(list_of_ints, []).success
    assert not eval_config_value_from_dagster_type(list_of_ints, [None]).success
    assert eval_config_value_from_dagster_type(list_of_ints, [1]).success

    nullable_list_of_ints = Nullable(List(Int))

    assert eval_config_value_from_dagster_type(nullable_list_of_ints, None).success
    assert eval_config_value_from_dagster_type(nullable_list_of_ints, []).success
    assert not eval_config_value_from_dagster_type(nullable_list_of_ints, [None]).success
    assert eval_config_value_from_dagster_type(nullable_list_of_ints, [1]).success

    list_of_nullable_ints = List(Nullable(Int))

    assert not eval_config_value_from_dagster_type(list_of_nullable_ints, None).success
    assert eval_config_value_from_dagster_type(list_of_nullable_ints, []).success
    assert eval_config_value_from_dagster_type(list_of_nullable_ints, [None]).success
    assert eval_config_value_from_dagster_type(list_of_nullable_ints, [1]).success

    nullable_list_of_nullable_ints = Nullable(List(Nullable(Int)))

    assert eval_config_value_from_dagster_type(nullable_list_of_nullable_ints, None).success
    assert eval_config_value_from_dagster_type(nullable_list_of_nullable_ints, []).success
    assert eval_config_value_from_dagster_type(nullable_list_of_nullable_ints, [None]).success
    assert eval_config_value_from_dagster_type(nullable_list_of_nullable_ints, [1]).success


def test_nullable_dict():
    dict_with_int = Dict({'int_field': Field(Int)})

    assert not eval_config_value_from_dagster_type(dict_with_int, None).success
    assert not eval_config_value_from_dagster_type(dict_with_int, {}).success
    assert not eval_config_value_from_dagster_type(dict_with_int, {'int_field': None}).success
    assert eval_config_value_from_dagster_type(dict_with_int, {'int_field': 1}).success

    nullable_dict_with_int = Nullable(Dict({'int_field': Field(Int)}))

    assert eval_config_value_from_dagster_type(nullable_dict_with_int, None).success
    assert not eval_config_value_from_dagster_type(nullable_dict_with_int, {}).success
    assert not eval_config_value_from_dagster_type(
        nullable_dict_with_int, {'int_field': None}
    ).success
    assert eval_config_value_from_dagster_type(nullable_dict_with_int, {'int_field': 1}).success

    dict_with_nullable_int = Dict({'int_field': Field(Nullable(Int))})

    assert not eval_config_value_from_dagster_type(dict_with_nullable_int, None).success
    assert not eval_config_value_from_dagster_type(dict_with_nullable_int, {}).success
    assert eval_config_value_from_dagster_type(dict_with_nullable_int, {'int_field': None}).success
    assert eval_config_value_from_dagster_type(dict_with_nullable_int, {'int_field': 1}).success

    nullable_dict_with_nullable_int = Nullable(Dict({'int_field': Field(Nullable(Int))}))

    assert eval_config_value_from_dagster_type(nullable_dict_with_nullable_int, None).success
    assert not eval_config_value_from_dagster_type(nullable_dict_with_nullable_int, {}).success
    assert eval_config_value_from_dagster_type(
        nullable_dict_with_nullable_int, {'int_field': None}
    ).success
    assert eval_config_value_from_dagster_type(
        nullable_dict_with_nullable_int, {'int_field': 1}
    ).success


def test_any_with_default_value():
    dict_with_any = Dict({'any_field': Field(Any, default_value='foo', is_optional=True)})
    result = eval_config_value_from_dagster_type(dict_with_any, None)
    assert result.success
    assert result.value == {'any_field': 'foo'}



# from bigco import PublicCloudConn, set_value_in_cloud_store

from dagster import (
    Dict,
    execute_pipeline,
    Field,
    InputDefinition,
    Int,
    OutputDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    ResourceDefinition,
    String,
    solid,
)


class PublicCloudConn:
    def __init__(self, username, password):
        self.username = username
        self.password = password


def set_value_in_cloud_store(_conn, _key, _value):
    # imagine this doing something
    pass


class PublicCloudStore:
    def __init__(self, username, password):
        # create credential and store it
        self.conn = PublicCloudConn(username, password)

    def record_value(self, log, key, value):
        log.info(
            'Setting key={key} value={value} in cloud'.format(
                key=key, value=value
            )
        )
        set_value_in_cloud_store(self.conn, key, value)


class InMemoryStore:
    def __init__(self):
        self.values = {}

    def record_value(self, log, key, value):
        log.info(
            'Setting key={key} value={value} in memory'.format(
                key=key, value=value
            )
        )
        self.values[key] = value


def define_in_memory_store_resource():
    return ResourceDefinition(
        resource_fn=lambda _: InMemoryStore(),
        description='''
    An in-memory key value store that requires no configuration. Useful for unittesting.
    ''',
    )


def define_cloud_store_resource():
    return ResourceDefinition(
        resource_fn=lambda info: PublicCloudStore(
            info.config['username'], info.config['password']
        ),
        config_field=Field(
            Dict({'username': Field(String), 'password': Field(String)})
        ),
        description='''This represents some cloud-hosted key value store.
        Username and password must be provided via configuration for this to
        work''',
    )


@solid(
    inputs=[InputDefinition('num_one', Int), InputDefinition('num_two', Int)],
    outputs=[OutputDefinition(Int)],
)
def add_ints(info, num_one, num_two):
    sum_ints = num_one + num_two
    info.resources.store.record_value(info.log, 'add', sum_ints)
    return sum_ints


def define_resource_test_pipeline():
    return PipelineDefinition(
        name='resource_test_pipeline',
        solids=[add_ints],
        context_definitions={
            'local': PipelineContextDefinition(
                resources={'store': define_in_memory_store_resource()}
            ),
            'cloud': PipelineContextDefinition(
                resources={'store': define_cloud_store_resource()}
            ),
        },
    )


if __name__ == '__main__':
    result = execute_pipeline(
        define_resource_test_pipeline(),
        environment={
            'context': {
                'cloud': {
                    'resources': {
                        'store': {
                            'config': {
                                'username': 'some_user',
                                'password': 'some_password',
                            }
                        }
                    }
                }
            },
            'solids': {
                'add_ints': {
                    'inputs': {
                        'num_one': {'value': 2},
                        'num_two': {'value': 6},
                    }
                }
            },
        },
    )

    result = execute_pipeline(
        define_resource_test_pipeline(),
        environment={
            'context': {'local': {}},
            'solids': {
                'add_ints': {
                    'inputs': {
                        'num_one': {'value': 2},
                        'num_two': {'value': 6},
                    }
                }
            },
        },
    )



from dagster import execute_pipeline
from dagster.tutorials.intro_tutorial.repos import define_repo


def test_define_repo():
    repo = define_repo()
    assert repo.name == 'demo_repository'
    assert repo.has_pipeline('repo_demo_pipeline')
    result = execute_pipeline(repo.get_pipeline('repo_demo_pipeline'))
    assert result.success



import pytest

from dagster import PipelineConfigEvaluationError, execute_pipeline
from dagster.tutorials.intro_tutorial.inputs import (
    execute_with_another_world,
    define_hello_typed_inputs_pipeline,
)
from dagster.tutorials.utils import check_cli_execute_file_pipeline
from dagster.utils import script_relative_path


def test_hello_inputs_parameterized_pipeline():
    result = execute_with_another_world()
    assert result.success
    solid_result = result.result_for_solid('add_hello_to_word')
    assert solid_result.transformed_value() == 'Hello, Mars!'


def test_hello_inputs_parameterized_cli_pipeline():
    check_cli_execute_file_pipeline(
        script_relative_path('../../../dagster/tutorials/intro_tutorial/inputs.py'),
        'define_hello_inputs_pipeline',
        script_relative_path('../../../dagster/tutorials/intro_tutorial/inputs_env.yml'),
    )


def test_hello_typed_inputs():
    with pytest.raises(
        PipelineConfigEvaluationError,
        match=(
            'Type failure at path '
            '"root:solids:add_hello_to_word_typed:inputs:word:value" on type "String"'
        ),
    ):
        execute_pipeline(
            define_hello_typed_inputs_pipeline(),
            {'solids': {'add_hello_to_word_typed': {'inputs': {'word': {'value': 343}}}}},
        )


def test_hello_typed_bad_structure():
    with pytest.raises(
        PipelineConfigEvaluationError,
        match=('Value for selector type String.InputSchema must be a dict got '),
    ):
        execute_pipeline(
            define_hello_typed_inputs_pipeline(),
            {'solids': {'add_hello_to_word_typed': {'inputs': {'word': {'Foobar Baz'}}}}},
        )


def test_hello_typed():
    result = execute_pipeline(
        define_hello_typed_inputs_pipeline(),
        {'solids': {'add_hello_to_word_typed': {'inputs': {'word': {'value': 'Foobar Baz'}}}}},
    )
    assert result.success
    assert len(result.result_list) == 1
    assert (
        result.result_for_solid('add_hello_to_word_typed').transformed_value()
        == 'Hello, Foobar Baz!'
    )



import uuid

from dagster.utils.test import create_test_runtime_execution_context

# pylint: disable=W0212


def test_noarg_ctor():
    context = create_test_runtime_execution_context()
    assert uuid.UUID(context.run_id)



import time

from dagster import (
    DependencyDefinition,
    InputDefinition,
    MultipleResults,
    OutputDefinition,
    PipelineDefinition,
    solid,
    SolidInstance,
)


def nonce_solid(name, n_inputs, n_outputs):
    """Creates a solid with the given number of (meaningless) inputs and outputs.

    Config controls the behavior of the nonce solid."""

    @solid(
        name=name,
        inputs=[
            InputDefinition(name='input_{}'.format(i)) for i in range(n_inputs)
        ],
        outputs=[
            OutputDefinition(name='output_{}'.format(i))
            for i in range(n_outputs)
        ],
    )
    def solid_fn(info, **_kwargs):
        for i in range(200):
            time.sleep(0.02)
            if i % 1000 == 420:
                info.log.error(
                    'Error message seq={i} from solid {name}'.format(
                        i=i, name=name
                    )
                )
            elif i % 100 == 0:
                info.log.warning(
                    'Warning message seq={i} from solid {name}'.format(
                        i=i, name=name
                    )
                )
            elif i % 10 == 0:
                info.log.info(
                    'Info message seq={i} from solid {name}'.format(
                        i=i, name=name
                    )
                )
            else:
                info.log.debug(
                    'Debug message seq={i} from solid {name}'.format(
                        i=i, name=name
                    )
                )
        return MultipleResults.from_dict(
            {'output_{}'.format(i): 'foo' for i in range(n_outputs)}
        )

    return solid_fn


def define_spew_pipeline():
    return PipelineDefinition(
        name='log_spew',
        solids=[
            nonce_solid('no_in_two_out', 0, 2),
            nonce_solid('one_in_one_out', 1, 1),
            nonce_solid('one_in_two_out', 1, 2),
            nonce_solid('two_in_one_out', 2, 1),
            nonce_solid('one_in_none_out', 1, 0),
        ],
        dependencies={
            SolidInstance('no_in_two_out', alias='solid_a'): {},
            SolidInstance('one_in_one_out', alias='solid_b'): {
                'input_0': DependencyDefinition('solid_a', 'output_0')
            },
            SolidInstance('one_in_two_out', alias='solid_c'): {
                'input_0': DependencyDefinition('solid_a', 'output_1')
            },
            SolidInstance('two_in_one_out', alias='solid_d'): {
                'input_0': DependencyDefinition('solid_b', 'output_0'),
                'input_1': DependencyDefinition('solid_c', 'output_0'),
            },
            SolidInstance('one_in_one_out', alias='solid_e'): {
                'input_0': DependencyDefinition('solid_c', 'output_0')
            },
            SolidInstance('two_in_one_out', alias='solid_f'): {
                'input_0': DependencyDefinition('solid_d', 'output_0'),
                'input_1': DependencyDefinition('solid_e', 'output_0'),
            },
            SolidInstance('one_in_none_out', alias='solid_g'): {
                'input_0': DependencyDefinition('solid_f', 'output_0')
            },
        },
    )
